#include <avr/io.h>
#include <avr/interrupt.h>
#include <inttypes.h>
#include <math.h>

#define SEGA 4
#define SEGB 2
#define SEGC 1
#define SEGD 32
#define SEGE 64
#define SEGF 16
#define SEGG 8
#define SEGP 128

uint8_t font[16] = {
    SEGA | SEGB | SEGC | SEGD | SEGE | SEGF,
    SEGB | SEGC,
    SEGA | SEGB | SEGD | SEGE | SEGG,
    SEGA | SEGB | SEGC | SEGD | SEGG,
    SEGB | SEGC | SEGF | SEGG,
    SEGA | SEGC | SEGD | SEGF | SEGG,
    SEGA | SEGC | SEGD | SEGE | SEGF | SEGG,
    SEGA | SEGB | SEGC,
    SEGA | SEGB | SEGC | SEGD | SEGE | SEGF | SEGG,
    SEGA | SEGB | SEGC | SEGD | SEGF | SEGG,
    SEGA | SEGB | SEGC | SEGE | SEGF | SEGG,
    SEGC | SEGD | SEGE | SEGF | SEGG,
    SEGA | SEGD | SEGE | SEGF,
    SEGB | SEGC | SEGD | SEGE | SEGG,
    SEGA | SEGD | SEGE | SEGF | SEGG,
    SEGA | SEGE | SEGF | SEGG,
};
/* LED */
uint8_t dsp[2] = {0, 0};
char leda = 0;
char ledc = 0;
/* Timer */
volatile int oticks = 0;
unsigned long mnow;
/* Pulse counter */
volatile char pstate = 0;
char pval = 0;
/* Switch */
volatile char sstate = 0;
int stime = 0;
/* Temp sensor */
volatile char tstate = 0;
volatile char tlock = 0;
unsigned long tstart;
unsigned long ttime;
unsigned long ttimea = 10000;
char tavgok = 0;
/* Conversion loop */
int tempk;
volatile ktok = 0;
/* Zero-cross detector*/
volatile char zok = 0;
volatile char ztime = 0;
/* Triac */
char trstate = 0;
char tron = 0;
volatile char trtime;
volatile char trdelay = 0;

void init(void)
{
    /* Timer init
     * Timer 0 cycles the Triac
     * Timer 1 is used for global timing
     * Timer 2 cycles the LED display
     */
    OCR0A = 100;
    TCCR0A = 2;
    TCCR0B = 1;
    TIMSK0 = 2;
    TCCR1A = 0;
    TCCR1B = 1;
    TIMSK1 = 1;
    OCR2A = 16;
    TCCR2A = 2;
    TCCR2B = 4;
    TIMSK2 = 2;
    
    /*
     * B0..2 = Pulse sensor
     * B3..5 = ISP
     * B6..7 = CLK
     */
    DDRB = 0x38;
    PORTB = 0x07;
    PCMSK0 = 0x07;
    PCICR = 0x01;
    /*
     * C0..5 = LEDA0..5
     * C6 = /RESET
     * C7 = NC
     */
    DDRC = 0x3f;
    PORTC = 0x00;
    /*
     * D0 = Triac
     * D1 = NTC FET
     * D2 = ZCD (INT0)
     * D3 = NTC Op-amp (INT1)
     * D4..5 = LEDA6..7
     * D6..7 = LEDC0..1
     */
    DDRD = 0xf3;
    PORTD = 0x00;
    EICRA = 0x0d;
    EIMSK = 0x03;
}

unsigned char bindisp(unsigned char num)
{
    unsigned char ret;
    
    ret = 0;
    if(num & 1)
	ret |= SEGA;
    if(num & 2)
	ret |= SEGB;
    if(num & 4)
	ret |= SEGC;
    if(num & 8)
	ret |= SEGD;
    if(num & 16)
	ret |= SEGE;
    if(num & 32)
	ret |= SEGF;
    if(num & 64)
	ret |= SEGG;
    if(num & 128)
	ret |= SEGP;
    return(ret);
}

void display(char num)
{
    dsp[0] = font[(num / 10) % 10];
    dsp[1] = font[num % 10];
}

void disphex(unsigned char num)
{
    dsp[0] = font[(num & 0xf0) >> 4];
    dsp[1] = font[num & 0x0f];
}

unsigned long getticks(void)
{
    uint16_t v;
    unsigned long r;
    
    cli();
    v = TCNT1;
    r = v + (((unsigned long)oticks) << 16);
    if((TIFR1 & 0x01) && !(v & 0x8000))
	r += 0x10000;
    sei();
    return(r);
}

void ledcycle(void)
{
    uint8_t c, d, v;
    
    if(++leda >= 8) {
	leda = 0;
	if(++ledc >= 2)
	    ledc = 0;
    }
    if(dsp[ledc] & (1 << leda)) {
	if(leda < 6) {
	    c = 1 << leda;
	    d = 0;
	} else {
	    c = 0;
	    d = 0x10 << (leda - 6);
	}
	d |= ledc?0x40:0x80;
    } else {
	c = d = 0;
    }
    PORTC = c;
    PORTD = (PORTD & 0x0f) | d;
}

void tempcycle(void)
{
    if(tstate == 0) {
	if((PIND & 8) && (tlock == 0)) {
	    cli();
	    PORTD |= 2;
	    sei();
	    tstart = mnow;
	    tstate = 1;
	}
    } else if(tstate == 1) {
	if(mnow - tstart > 1000) {
	    cli();
	    PORTD &= ~2;
	    sei();
	    tstate = 0;
	    tstart = mnow;
	}
    }
}

void calcavg(void)
{
    if(tlock == 1) {
	tlock = 2;
	ttimea = ((ttimea * 15) + ttime) >> 4;
	tlock = 0;
	tavgok = 1;
    }
}

void convcycle(void)
{
    static char state = 0;
    static unsigned long last = 0;
    static float a, ra, l, t;
    
    /*
     * Theoretically:
     *  t = RC * ln(2) => R = t / (C * ln(2))
     *  R = A * exp(B / T) => T = B / ln(R / A)
     *  T = B / ln(R / (A * C * ln(2)))
     * In the following: 
     *  a = ttimea as float
     *  C = 1e6 / (A * C * ln(2))
     *  ra = a * C
     *  l = ln(ra)
     *  t = B / l
     * Note, temperature is in Kelvin
     */
#define C 9.792934
#define B 4020.0
    if(state == 0) {
	if((mnow - last > 200000) && tavgok) {
	    a = (float)ttimea;
	    state = 1;
	    tavgok = 0;
	    last = mnow;
	}
    } else if(state == 1) {
	ra = a * C;
	state = 2;
    } else if(state == 2) {
	l = log(ra);
	state = 3;
    } else if(state == 3) {
	t = B / l;
	state = 4;
    } else if(state == 4) {
	tempk = (int)t;
	ktok = 1;
	state = 0;
    }
}

int main(void)
{
    int state, cur, run, rstate, delta;
    unsigned long utime;
    
    state = 0;
    cur = 100;
    run = 0;
    rstate = 0;
    init();
    sei();
    display(0);

    while(1) {
	mnow = getticks();
	tempcycle();
	calcavg();
	convcycle();

#if 1
	/*
	 * User interface
	 */
	if(state == 0) {
	    /* Display temperature */
	    if(ktok) {
		ktok = 0;
		if((tempk >= 273) && (tempk <= 372)) {
		    display(tempk - 273);
		} else {
		    dsp[0] = dsp[1] = SEGG;
		}
	    }
	    if(pval != 0)
		state = 1;
	    if(sstate == 2) {
		sstate = 0;
		if(stime > 10)
		    state = 2;
		else
		    run = !run;
	    }
	    if(run)
		dsp[1] |= SEGP;
	    else
		dsp[1] &= ~SEGP;
	} else if(state == 1) {
	    /* Temp setting */
	    if(pval != 0) {
		cur += pval;
		pval = 0;
		if(cur < 0)
		    cur = 0;
		if(cur > 100)
		    cur = 100;
		if(cur < 100)
		    display(cur);
		else
		    dsp[0] = dsp[1] = SEGG;
		utime = mnow;
	    }
	    if(mnow - utime > 1000000)
		state = 0;
	    if(sstate == 2) {
		run = !run;
		sstate = 0;
	    }
	} else if(state == 2) {
	    /* Display raw temp time reading */
	    if(ttimea < 20000) {
		display((ttimea / 100) % 100);
		dsp[0] |= SEGP;
		if(ttimea >= 10000)
		    dsp[1] |= SEGP;
	    } else {
		display(ttimea / 1000);
	    }
	    if(sstate == 2) {
		state = 0;
		sstate = 0;
	    }
	}
	/*
	 * Set Triac to match temperature
	 */
	if(run) {
	    delta = cur - (tempk - 273);
	    if(rstate == 0) {
		if(delta > 0) {
		    tron = 1;
		    if(delta > 8) {
			/* For some reason, the Triac currently doesn't
			 * trigger on one of the AC half-cycles below 0.7
			 * ms. */
			trdelay = 7;
		    } else {
			trdelay = 79 - (delta * 9);
		    }
		} else {
		    tron = 0;
		    rstate = 1;
		}
	    } else if(rstate == 1) {
		tron = 0;
		if(delta >= 2)
		    rstate = 0;
	    }
	} else {
	    tron = 0;
	}
#endif
	/*
	  dsp[0] = bindisp((ttimea & 0xff00) >> 8);
	  dsp[1] = bindisp(ttimea & 0x00ff);
	*/
	/*
	  disphex((ttimea & 0xff000) >> 12);
	*/
#if 0
	/*
	  Temp display
	*/
	if(ttimea < 20000) {
	    display((ttimea / 100) % 100);
	    dsp[0] |= SEGP;
	    if(ttimea >= 10000)
		dsp[1] |= SEGP;
	} else {
	    display(ttimea / 1000);
	}
#endif
#if 0
	/*
	 * ZVD debug
	 */
	if(zok) {
	    if(++cur > 99)
		cur = 0;
	    display(cur);
	    zok = 0;
	}
#endif
#if 0
	/*
	  Phony Triac control
	 */
	if(pval != 0) {
	    cur += pval;
	    if(cur < 0)
		cur = 0;
	    if(cur > 99)
		cur = 99;
	    display(cur);
	    trdelay = 99 - cur;
	    pval = 0;
	}
	if(sstate == 2) {
	    tron = !tron;
	    sstate = 0;
	}
	if(tron)
	    dsp[1] |= SEGP;
	else
	    dsp[1] &= ~SEGP;
#endif
#if 0
	/*
	  Pulse counter display
	*/
	cur += pval;
	pval = 0;
	if(sstate == 2) {
	    cur = stime;
	    sstate = 0;
	}
	if(cur > 99)
	    cur = 99;
	if(cur < -99)
	    cur = -99;
	if(cur < 0) {
	    display(-cur);
	    dsp[0] |= SEGP;
	} else {
	    display(cur);
	}
	if(PINB & 4)
	    dsp[1] |= SEGP;
#endif
    }
}

ISR(SIG_INTERRUPT0)
{
    ztime = 0;
    zok = 1;
}

ISR(SIG_INTERRUPT1)
{
    unsigned long now;
    
    now = getticks();
    if(tstate == 0) {
	tstate = 1;
	if(tlock != 2)
	    ttime = now - tstart;
	tstart = now;
	PORTD |= 2;
	tlock = 1;
    }
}

ISR(SIG_OUTPUT_COMPARE0A)
{
    if(trstate == 0) {
	ztime++;
	if(tron && (ztime >= trdelay)) {
	    PORTD |= 1;
	    trstate = 1;
	    trtime = 0;
	}
    } else if(trstate == 1) {
	trtime++;
	if(trtime >= 5) {
	    PORTD &= ~1;
	    trstate = 0;
	}
    }
}

ISR(SIG_OUTPUT_COMPARE2A)
{
    ledcycle();
}

ISR(SIG_OVERFLOW1)
{
    oticks++;
}

ISR(SIG_PIN_CHANGE0)
{
    if((sstate == 0) && !(PINB & 4)) {
	stime = oticks;
	sstate = 1;
    }
    if((sstate == 1) && (PINB & 4)) {
	stime = oticks - stime;
	sstate = 2;
    }
    if(pstate == 0) {
	if((PINB & 2) == 0) {
	    pstate = 1;
	} else if((PINB & 1) == 0) {
	    pstate = 2;
	}
    } else if(pstate == 1) {
	if((PINB & 1) == 0) {
	    pval++;
	    pstate = 3;
	} else {
	    pstate = 0;
	}
    } else if(pstate == 2) {
	if((PINB & 2) == 0) {
	    pval--;
	    pstate = 3;
	} else {
	    pstate = 0;
	}
    } else {
	if((PINB & 2) && (PINB & 1))
	    pstate = 0;
    }
}
